
-----
-- += -= *= /= ..= token filter                          Copyright (c) 2006 Asko Kauppi
--
-- Syntactic sugar, especially useful for longer variable and/or structure field names
--
--      name += exp     ->  name= name + exp
--      name -= exp     ->  name= name - exp
--      name *= exp     ->  name= name * exp
--      name /= exp     ->  name= name / exp
--      name ..= exp    ->  name= name .. exp
--
-- Usage:
--      lua -lftools -finc ...
--
-- Requires:    Lua 5.1, with token filter patch applied
--              ftools.lua
--
-- Note: 
--      The nice part about this sugar is, we don't have to care of the right hand side
--      (exp) at all (luckily, it would be VERY hard to do!). It's enough to expand the
--      left hand side "name op=" -> "name= name op" and let Lua continue from there. :)
-----

if type( rawget(_G,"FILTER") )~="table" then
    error "Requires ftools.lua to be loaded"
end

local match_op= { ['+']=1, ['-']=1, ['*']=1, ['/']=1, [".."]=1 }

-- sugar: "<name> { ('.' <name>)|('[' exp ']')} op '='" --> "<name> {'.' <name>} '=' <same-as-left-side> op"
--
local function INC(p)
    p["until"]( p, "<name>" )   -- p:until "<name>"

    local n=1
    local op

    -- Gather any '.' <name> into the same group
    --
    while true do
        op= p:peek()
        if op=='.' then
            local _,tk2= p:read(2)
            -- allow both names and reserved words, this will allow "t.until += 1" if some
            -- other filter makes sure reserved words are translated appropriately.
            -- 
            if tk2 == "<name>" or string.sub(tk2,1,1) then
                n=n+2
            else
                n=0; break   -- '.' <number> or something else querky - give up!
            end
        elseif op=='[' then
            local brackets= 0
            repeat
                local tk= p:read()  -- first one is the '['
                if tk=='[' then brackets= brackets+1
                elseif tk==']' then brackets= brackets-1
                end
                n= n+1
            until brackets==0
        else
            break
        end
    end

    if n>0 and match_op[op] and p:peek()=='=' then
        -- We've got what we wanted!
        --
        -- p:repeat() will do a duplicate on the statement, since we carefully did NOT read in
        -- operator, and '=' (yet)
        -- 
        p:replace( 0, '=', p["repeat"]( p, n ) )
        p:read(2)   -- op '='
        p:replace( 1 )  -- yam! (eat the '=' out)
    end
end

FILTER.link( INC )
