--
-- ftools filter
--
-- Tools for making token filters.
--
-- Usage:
--      lua -lftools -lfmy1 -lfmy2 app.lua
--          ^^^^^^^^
--
-- In other words, load this FIRST, then any filters actually using these services. This chunk
-- has the FILTER routine (actually, defined as a callable table), which is the entry point
-- patched Lua is looking for.
--
-- Other filters join in by calling FILTER.join(), forming a pipe of filters at load time.
--
-- Copyright (c) 2006 Asko Kauppi <akauppi@gmail.com>
--

local DEBUG=
        function() end or   -- comment this line to enable debug
        function( me, ... )
            if true then
                io.write( string.rep("   ",me._level) )
                print(...)
            end
        end

local DEBUG_CACHE= 
        function() end or   -- comment this line to enable debug
        function(me)
            io.write "\t\t"
            for _,mm in ipairs(me) do
                io.write( "{ "..mm[1].." "..mm[2]..(mm[3] and " "..mm[3] or "").." }, " )
            end
            print ""
        end

-----
-- {line_num, tk_str, [val] }= raw_read( me )
--
-- Read one token from the token flow.
--
local function raw_read( me )
    local ln, tk, val= me._get()
DEBUG( me, "get:", tk.." "..(val or "\t"), ln )
    return { ln,tk,val }
end

-----
-- tk [,val] [, ...]= read( me [,n=1] )
--
-- Read N tokens, returning each, and placing their info on the cache.
-- Use 'forget()' to eat up the tokens, or 'flush()' to pass them on.
--
local function read( me, n )
    -- Handle peeked items first ([-1..-N])    
    local m= me[-1]
    if m then
        local k=-1
        repeat me[k]= me[k-1]; k=k-1 until me[k]==nil
    else
        m= raw_read(me)
    end
    me[#me+1]= m
    local ln,tk,val= unpack(m)

    ------
    -- TBD: Lua bug?
    --
    -- For some reason, we're getting tokens ('make test-inc', lines 10-11) backwards.
    -- May be a Lua 5.1.1 or token filter bug - we're catering for it elsewhere, important
    -- since wrong numbering will cause "function ()" not to compile if name and paranthesis
    -- are on wrong lines
    --
    if true then
        if me._last_read_line and me._last_read_line > ln then
            --error( "Weird: getting line numbers backwards!!! ".. me._last_read_line.." -> "..ln )

            -- compensate: cause latest read line to be >= previous ones
            --
           ln= me._last_read_line
            me[#me][1]= ln
        end
    end
    me._last_read_line= ln
    -----

DEBUG_CACHE(me)
    if n and n>1 then
        if val then return tk,val, read( me, n-1 )
        else return tk, read( me, n-1 )
        end
    else
        -- Make sure unprocessed tokens have been passed on, even if the customer
        -- filter didn't.
        --
        if tk=="<eof>" then
            me:flush()
        end
        return tk,val
    end
end

-----
-- tk [,val] [, ...]= peek( me [,n=1] )
--
-- Like 'read()' but leaves the insertion place (for 'replace()' before the tokens,
-- as opposed to after them).
--
local function peek( me, n )
    local k= -1
    while me[k] do k=k-1 end  -- already peek'ed items

    me[k]= raw_read(me)
    
    local _,tk,val= unpack( me[k] )
    if n and n>1 then
        if val then return tk,val, peek(me,n-1)
        else return tk, peek(me,n-1)
        end
    else
        return tk,val   -- val may be nil
    end
end

-----
-- tk [,val]= until( me, tk_str )
--
-- Reads tokens, passing them through, until 'tk' is met.
-- 'tk' is left in the cache, so caller may call 'forget()' or 'flush()'.
--
-- For value tokens, s.a. "<number>" only the token name is given, it will
-- match a token of that kind, regardless of value, and return both.
-- 
local function until_( me, tk2 )
    local tk,val
    while true do
        tk,val= me:read()
        if tk==tk2 then break end
        me:flush()     -- pass it on
    end
    return tk,val
end


-----
-- void= replace( me, n_int, [tk_str [,val_str/num] [, ...]] )
--
-- Removes 'n' (>=0) tokens from the cache, replacing them with the given ones.
--
local function replace( me, n, ... )
    local dot_n= 1
    local k= #me-n+1

    if k<1 then
        error "FILTER ERROR: not so many tokens in the cache!"
    end

    local ln
    if #me>0 then
        ln= (me[k] or me[#me])[1]    -- line that is being replaced, or last line ('n'==0)
    else
        ln= me._last_flushed_line or 1
    end

    while true do
        local tk,val
        tk= (select( dot_n, ... ))
        dot_n= dot_n+1
        if tk=="<number>" or tk=="<string>" or tk=="<name>" then
            val= select( dot_n, ... )
            dot_n = dot_n+1
            if not val then
                error "FILTER ERROR: token without a value!"
            end
        end
        if tk then
            me[k]= { ln, tk, val }
DEBUG_CACHE(me)
        else
            while #me>k-1 do me[#me]=nil end    -- remove the tail
            break
        end
        k= k+1
    end
end

-----
-- flush()
--
local function flush( me )
    for i,v in ipairs(me) do   -- 1..N
DEBUG_CACHE(me)
        -----
        -- Lua gets really p**sed if line numbers go backwards; it should not happen!
        if me._last_flushed_line and v[1] < me._last_flushed_line then
            error( "Writing tokens with backwards line numbers! "..(me._last_flushed_line).." "..v[1] )
        end
        -----
DEBUG( me, "put:", v[2].." "..(v[3] or "\t"), v[1] )
        me._put( unpack(v) )
        me[i]= nil  -- this IS allowed (messing with 'i' is not)
        me._last_flushed_line= v[1]  -- remember in case 'replace(0,...)'
    end
end

-----
-- ...= repeat(n)
--
-- Repeat the last 'n' tokens, can be used in conjunction with 'replace()' to duplicate tokens
--
local function repeat_( me, n )
    n= n or 1
    if n > #me then
        error "FILTER ERROR: not enough tokens to repeat!"
    end
    local _,tk,val= unpack( me[#me-(n-1)] )
    if n>1 then
        if val then return tk,val, repeat_( me, n-1 )
        else return tk, repeat_( me, n-1 )
        end
    else
        return tk,val    -- val may be nil
    end
end

-----
-- new()
--
local methods= {
    read=      read,
    peek=      peek,
    flush=     flush, 
    replace=   replace,
    ["until"]= until_,    -- 5.2: could reserved names be allowed as table indices? They can.
    ["repeat"]= repeat_,
}

local counter=0

local function new( get, put )
    local me= { _last_flushed_line=nil,
                _get=get, _put=put }
    me._level= counter
    counter= counter+1
    setmetatable( me, { __index=methods } )
    return me
end


--
assert( not rawget(_G,"FILTER"), "FILTER already defined -- aborting!" )

--
local function pipe(f,get,put)
	put = put or coroutine.yield
	local F=coroutine.wrap(f)
	F(get,put)     -- spins into the filter, 'put "spinback"' takes back
	return F
end

local FILTER_FUNC

local filters= {}   

-----
-- void= FILTER( get_func, source_str ) get:    line,token[,value]= get()
--                                      source: "=stdin" or "@file.lua"
--
-- line,token[,value]= FILTER()         Lua requesting a token (lots of these!)
--
-- void= FILTER.link( func(obj) )       customer filter tying into us
--
rawset( _G, "FILTER", {
  __call= function( _, get, source )
    if get then                  -- get,source
        -- If we set FILTER_FUNC= get, we lose the filters already set up
        -- If we don't set, things don't seem to work (= get is different)
        --
        FILTER_FUNC= get
        for _,g in ipairs(filters) do
            FILTER_FUNC= pipe( g, FILTER_FUNC )
        end
    else                         -- nil,nil: squeezing a token
        return FILTER_FUNC()
    end
  end,
  
  -- Set up a new filter function into the chain
  --
  link= function( f )
        assert( type(f)=="function", "Usage: FILTER.link( func(obj) )" )

        local g= function(get,put)
                    local me= new(get,put)
                    put ""  -- dummy, isn't it?
                    while true do f(me); me:flush() end     -- flushing is good for cache
                 end
       assert( FILTER_FUNC )
	   FILTER_FUNC= pipe( g, FILTER_FUNC )

       -- Remember the functions
	   filters[ #filters+1 ]= g
    end
} )

setmetatable( FILTER, FILTER )

